from typing import Callable, Any, Iterable, Awaitable, TypeVar

import asyncio
from functools import wraps

import tqdm  # type: ignore

from .fmap import fmap  # NOQA


A = TypeVar('A')


def as_submitted(it : Iterable[Awaitable[A]]
                 ) -> Iterable[Awaitable[A]]:
    _ = it
    _ = [asyncio.ensure_future(el) for el in _]

    return _


async def pure(el : A,
               ) -> A:
    return el


def run(coro : Awaitable[A],
        ) -> A:
    return asyncio\
        .get_event_loop()\
        .run_until_complete(coro)


def update_pbar(f    : Callable,
                pbar : tqdm.tqdm,
                n    : int = 1,
                ) -> Callable:
    @wraps(f)
    def _update_pbar(*args    : Any,
                     **kwargs : Any,
                     ) -> Any:
        result = f(*args, **kwargs)
        pbar.update(n)
        return result

    return _update_pbar



from pll import Pll


def test_doesnt_break():
    _ = Pll\
        .from_non_coroutines(range(10))\
        .map(lambda x: x ** 2)\
        .as_completed()\
        .aggregate(set)

    assert _ == set(x ** 2 for x in range(10))



from typing import (
    Generic, TypeVar, Iterable, Iterator,
    Awaitable, Callable, Any, List, Optional
)
from asyncio import AbstractEventLoop
from concurrent.futures import Executor

import asyncio
from functools import reduce

import tqdm  # type: ignore

from .util import as_submitted, pure, run, fmap, update_pbar


A = TypeVar('A')
B = TypeVar('B')


class Pll(Generic[A], Iterable[A]):
    def __init__(self,
                 it : Iterable[Awaitable[A]],
                 *,
                 executor : Executor = None,
                 pbars : List[tqdm.tqdm] = [],
                 ) -> None:
        self.it = it
        self.executor = executor
        self.pbars = pbars

    @classmethod
    def from_non_coroutines(cls,
                            it : Iterable[A],
                            ) -> 'Pll[A]':
        return cls(as_submitted(pure(el) for el in it))

    def __reconstruct(self,
                      it : Iterable[Awaitable[B]],
                      executor : Executor = None,
                      pbars : List[tqdm.tqdm] = None,
                      ) -> 'Pll[B]':
        if executor is None:
            executor = self.executor
        if pbars is None:
            pbars = self.pbars

        return Pll(it,
                   executor=executor,
                   pbars=pbars,
                   )

    def __iter__(self,
                 ) -> Iterator[A]:
        try:
            for el in self.it:
                yield run(el)
        finally:
            for pbar in self.pbars:
                pbar.close()

    def map(self,
            f : Callable[[A], B],
            ) -> 'Pll[B]':
        _ : Any
        _ = self.it
        _ = fmap(fmap(f), _)

        return self.__reconstruct(_)

    def map_in_executor(self,
                        f : Callable[[A], B],
                        executor : Executor = None,
                        loop : AbstractEventLoop = None,
                        ) -> 'Pll[B]':
        if executor is None:
            executor = self.executor
        if loop is None:
            loop = asyncio.get_event_loop()

        async def wrapped(coro : Awaitable[A]) -> B:
            return await loop.run_in_executor(executor, f, await coro)

        _ : Any
        _ = self.it
        _ = fmap(wrapped, _)
        _ = as_submitted(_)

        return self.__reconstruct(_)

    def bind(self,
             f : Callable[[A], Awaitable[B]],
             ) -> 'Pll[B]':
        async def wrapped(coro : Awaitable[A]) -> B:
            return await f(await coro)
        _ : Any
        _ = self.it
        _ = fmap(wrapped, _)
        _ = as_submitted(_)

        return self.__reconstruct(_)

    def show_progress(self,
                      style : Optional[str] = None,
                      ) -> 'Pll[A]':
        if style:
            _tqdm = getattr(tqdm, 'tqdm_' + style)
        else:
            _tqdm = tqdm.tqdm

        _ : Any
        _ = self.it
        _ = list(_)
        pbar = _tqdm(total=len(_))
        _ = fmap(fmap(update_pbar(lambda x: x, pbar)), _)

        return self.__reconstruct(_,
                                  pbars=[*self.pbars, pbar],
                                  )

    def as_completed(self,
                     ) -> 'Pll[A]':
        _ : Any
        _ = self.it
        _ = list(_)
        _ = as_submitted(asyncio.as_completed(_))  # type: ignore

        return self.__reconstruct(_)

    # def accumulate(self,
    #                f :,
    #                ):
    #     ...

    def reduce(self,
               f : Callable[[B, A], B],
               *args : B,
               ) -> B:
        return reduce(f, iter(self), *args)

    def aggregate(self,
                  f : Callable[[Iterable[A]], B],
                  ) -> B:
        return f(iter(self))



from typing import (
    Coroutine, Generator, Iterator, Callable,
    TypeVar, Any,
)
from asyncio import Task

import asyncio
from functools import partial, singledispatch


A = TypeVar('A')
B = TypeVar('B')


def compose(f : Callable[[A], B],
            g : Callable[..., A] = None,
            ) -> Callable[..., B]:
    if not g:
        return partial(compose, f)
    return lambda *args, **kwargs: f(g(*args, **kwargs))


def fmap(f   : Callable,
         obj : Any = None,
         ) -> Any:
    if obj is None:
        return partial(fmap, f)
    return _fmap(obj, f)


@singledispatch
def _fmap(obj, f):  # type: ignore
    return obj.fmap(f)


@_fmap.register(list)
def _(obj, f):
    return [f(el) for el in obj]


@_fmap.register(tuple)  # type: ignore
def _(obj, f):
    return tuple(f(el) for el in obj)


@_fmap.register(Coroutine)  # type: ignore
def _(obj, f):
    async def wrapped():
        return f(await obj)
    return wrapped()


@_fmap.register(Task)  # type: ignore
def _(obj, f):
    async def wrapped():
        return f(await obj)
    return asyncio.ensure_future(wrapped())


@_fmap.register(Generator)  # type: ignore
def _(obj, f):
    return (f(el) for el in obj)


@_fmap.register(Iterator)  # type: ignore
def _(obj, f):
    return map(f, obj)


@_fmap.register(Callable)  # type: ignore
def _(obj, f):
    return compose(f, obj)



from setuptools import setup


setup(name='pll',
      version='0.1.1',
      author='Beta Faccion',
      author_email='betafcc@gmail.com',
      packages=['pll'],
      install_requires=['tqdm'],
      url='https://github.com/betafcc/pll',
      download_url='https://github.com/betafcc/pll/tarball/v0.1.1',
      )
